import multiprocessing
import os
import re
import time
from unittest.mock import AsyncMock, Mock, call, patch

import pytest

from upsies import errors
from upsies.jobs.torrent import CreateTorrentJob, _torrent_process
from upsies.utils import torrent
from upsies.utils.daemon import MsgType


@pytest.fixture
def tracker():
    tracker = Mock()
    tracker.configure_mock(
        name='AsdF',
        options={
            'source': 'AsdF',
            'randomize_infohash': 'maybe randomize infohash',
            'exclude': ('a', 'b'),
        },
        login=AsyncMock(),
        get_announce_url=AsyncMock(),
        logout=AsyncMock(),
        calculate_piece_size_=Mock(name='calculate_piece_size'),
        calculate_piece_size_min_max=Mock(name='calculate_piece_size_min_max'),
    )
    return tracker


def make_CreateTorrentProgress(**kwargs):
    kw = {
        'bytes_per_second': 'mock bytes_per_second',
        'filepath': 'mock filepath',
        'percent_done': 'mock percent_done',
        'piece_size': 'mock piece_size',
        'pieces_done': 'mock pieces_done',
        'pieces_total': 'mock pieces_total',
        'seconds_elapsed': 'mock seconds_elapsed',
        'seconds_remaining': 'mock seconds_remaining',
        'seconds_total': 'mock seconds_total',
        'time_finished': 'mock time_finished',
        'time_started': 'mock time_started',
        'total_size': 'mock total_size',
    }
    kw.update(kwargs)
    return torrent.CreateTorrentProgress(**kw)

def make_FindTorrentProgress(**kwargs):
    kw = {
        'exception': 'mock exception',
        'filepath': 'mock filepath',
        'files_done': 'mock files_done',
        'files_per_second': 'mock files_per_second',
        'files_total': 'mock files_total',
        'percent_done': 'mock percent_done',
        'seconds_elapsed': 'mock seconds_elapsed',
        'seconds_remaining': 'mock seconds_remaining',
        'seconds_total': 'mock seconds_total',
        'status': 'mock status',
        'time_finished': 'mock time_finished',
        'time_started': 'mock time_started',
    }
    kw.update(kwargs)
    return torrent.FindTorrentProgress(**kw)


def test_CreateTorrentJob_cache_id(tmp_path, tracker):
    job = CreateTorrentJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        ignore_cache=False,
        content_path='path/to/foo',
        tracker=tracker,
    )
    assert job.cache_id is None


def test_CreateTorrentJob_initialize(tracker, tmp_path):
    job = CreateTorrentJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        ignore_cache=False,
        content_path='path/to/foo',
        reuse_torrent_path='path/to/existing.torrent',
        tracker=tracker,
    )
    assert job._content_path == 'path/to/foo'
    assert job._torrent_path == f'{tmp_path / "foo"}.asdf.torrent'
    assert job._exclude_files == list(tracker.options['exclude'])
    assert job._reuse_torrent_path == 'path/to/existing.torrent'
    assert job.info == ''
    assert job._torrent_process is None


@pytest.mark.parametrize(
    argnames='exclude_defaults, exclude_files, exp_exclude_files',
    argvalues=(
        (
            [r'.*\.txt$'],
            ['*.jpg'],
            [r'.*\.txt$', '*.jpg'],
        ),
    ),
    ids=lambda v: repr(v),
)
def test_CreateTorrentJob_initialize_with_exclude_files(exclude_defaults, exclude_files, exp_exclude_files,
                                                        tracker, tmp_path, mocker):
    mocker.patch.dict(tracker.options, {'exclude': exclude_defaults})
    job = CreateTorrentJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        ignore_cache=False,
        content_path='path/to/foo',
        tracker=tracker,
        exclude_files=exclude_files,
    )
    assert job._exclude_files == list(exp_exclude_files)


@pytest.fixture
def job(tmp_path, tracker):
    return CreateTorrentJob(
        home_directory=tmp_path,
        cache_directory=tmp_path,
        ignore_cache=False,
        content_path='path/to/foo',
        reuse_torrent_path='path/to/existing.torrent',
        tracker=tracker,
    )


@pytest.mark.asyncio
async def test_CreateTorrentJob_activity_property(job, mocker):
    mock_activity = Mock()
    mocker.patch.object(job, '_activity', mock_activity)
    assert job.activity is mock_activity


@pytest.mark.asyncio
async def test_CreateTorrentJob_execute(job, mocker):
    mocker.patch.object(job, '_get_announce_url', AsyncMock())
    mocker.patch.object(job, '_handle_torrent_created', Mock())
    mocker.patch.object(job, '_start_torrent_creation_process', Mock())

    job.execute()
    await job._get_announce_url_task

    assert job._get_announce_url.call_args_list == [call()]
    assert job._handle_torrent_created.call_args_list == []
    assert job._start_torrent_creation_process.call_args_list == [call(job._get_announce_url.return_value)]
    assert not job.is_finished


@pytest.mark.asyncio
async def test_CreateTorrentJob_get_announce_url(job, mocker):
    mocks = AsyncMock(
        get_announce_url=AsyncMock(return_value='http://announce.url'),
        announce_url_callback=Mock(),
    )
    mocker.patch.object(job._tracker, 'get_announce_url', mocks.get_announce_url)

    def check_activity(*args, **kwargs):
        assert job.activity == 'announce_url'

    mocks.announce_url_callback.side_effect = check_activity
    job.signal.register('announce_url', mocks.announce_url_callback)

    assert job.activity == ''
    announce_url = await job._get_announce_url()
    assert announce_url == mocks.get_announce_url.return_value
    assert mocks.mock_calls == [
        call.announce_url_callback(Ellipsis),
        call.get_announce_url(),
        call.announce_url_callback('http://announce.url'),
    ]
    assert job.activity == ''

@pytest.mark.asyncio
async def test_CreateTorrentJob_get_announce_url_catches_RequestError_from_get_announce_url(job, mocker):
    mocks = AsyncMock(
        get_announce_url=AsyncMock(side_effect=errors.RequestError('no url found')),
        announce_url_callback=Mock(),
    )
    mocker.patch.object(job._tracker, 'get_announce_url', mocks.get_announce_url)
    job.signal.register('announce_url', mocks.announce_url_callback)

    assert job.activity == ''
    announce_url = await job._get_announce_url()
    assert announce_url is None
    assert mocks.mock_calls == [
        call.announce_url_callback(Ellipsis),
        call.get_announce_url(),
    ]
    assert job.errors == (errors.RequestError('no url found'),)
    assert job.is_finished
    assert job.activity == ''


def test_CreateTorrentJob_start_torrent_creation_process(job, mocker):
    DaemonProcess_mock = mocker.patch('upsies.utils.daemon.DaemonProcess', Mock(
        return_value=Mock(join=AsyncMock()),
    ))
    announce_url = 'http:/foo/announce'
    job._start_torrent_creation_process(announce_url)
    assert DaemonProcess_mock.call_args_list == [call(
        name=job.name,
        target=_torrent_process,
        kwargs={
            'content_path': 'path/to/foo',
            'announce': announce_url,
            'source': job._tracker.options['source'],
            'torrent_path': os.path.join(
                job.home_directory,
                f'foo.{job._tracker.options["source"].lower()}.torrent',
            ),
            'exclude': job._exclude_files,
            'use_cache': not job.ignore_cache,
            'reuse_torrent_path': job._reuse_torrent_path,
            'randomize_infohash': job._tracker.options['randomize_infohash'],
            'piece_size_calculator': job._tracker.calculate_piece_size,
            'piece_size_min_max_calculator': job._tracker.calculate_piece_size_min_max,
        },
        init_callback=job._handle_files,
        info_callback=job._handle_info_update,
        error_callback=job._handle_error,
        result_callback=job._handle_torrent_created,
        finished_callback=job.finish,
    )]
    assert job._torrent_process.start.call_args_list == [call()]


@pytest.mark.parametrize('torrent_process', (None, Mock()))
def test_CreateTorrentJob_finish(torrent_process, job):
    job._torrent_process = torrent_process
    assert not job.is_finished
    job.finish()
    assert job.is_finished
    if torrent_process:
        assert job._torrent_process.stop.call_args_list == [call()]
    else:
        assert job._torrent_process is None


@pytest.mark.parametrize('activity', ('', 'searching', 'hashing'))
@pytest.mark.parametrize('torrent_process', (None, 'mock torrent_process'), ids=lambda v: repr(v))
def test_CreateTorrentJob_cancel_search(torrent_process, activity, job, mocker):
    mocker.patch.object(job, '_torrent_process', (Mock() if torrent_process else None))
    mocker.patch.object(job, '_activity', activity)
    job.cancel_search()
    if activity == 'searching' and torrent_process:
        assert job._torrent_process.send.call_args_list == [call(MsgType.terminate, torrent.SKIP_SEARCHING)]
    elif torrent_process:
        assert job._torrent_process.send.call_args_list == []


def test_CreateTorrentJob_handle_files(job, mocker):
    mocks = Mock()
    job.signal.register('file_tree', mocks.tree)
    job.signal.register('file_list', mocks.list)
    mock_files = Mock(tree='mock tree', list='mock list')
    job._handle_files(mock_files)
    assert mocks.mock_calls == [
        call.tree('mock tree'),
        call.list('mock list'),
    ]


@pytest.mark.parametrize(
    argnames='progress, exp_exception, exp_activity',
    argvalues=(
        (make_CreateTorrentProgress(), None, 'hashing'),
        (make_FindTorrentProgress(status='verifying'), None, 'verifying'),
        (make_FindTorrentProgress(status='hit'), None, ''),
        (make_FindTorrentProgress(status='<anything else>'), None, 'searching'),
        (
            'invalid progress object',
            RuntimeError('Unexpected info update: {info!r}'),
            '',
        ),
    ),
    ids=lambda v: type(v).__name__,
)
def test_CreateTorrentJob_handle_info_update(progress, exp_exception, exp_activity, job, mocker):
    mocker.patch.object(job.signal, 'emit')
    assert job.activity == ''
    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            job._handle_info_update(progress)
        assert job.signal.emit.call_args_list == []
    else:
        job._handle_info_update(progress)
        assert job.signal.emit.call_args_list == [call('progress_update', progress)]
    assert job.activity == exp_activity


@pytest.mark.parametrize(
    argnames='torrent_path, exp_output',
    argvalues=(
        (None, ()),
        ('', ()),
        ('foo/bar.torrent', ('foo/bar.torrent',),),
    ),
)
def test_CreateTorrentJob_handle_torrent_created(torrent_path, exp_output, job):
    assert job.output == ()
    job._handle_torrent_created(torrent_path)
    assert job.output == exp_output


def test_CreateTorrentJob_handle_error(job, mocker):
    mocker.patch.object(job, 'exception')
    mocker.patch.object(job, 'error')
    job._handle_error('message')
    assert job.error.call_args_list == [call('message')]
    job._handle_error(errors.RequestError('message'))
    assert job.exception.call_args_list == [call(errors.RequestError('message'))]


class Callable:
    def __eq__(self, other):
        return callable(other)


@pytest.fixture
def queues():
    queues = Mock(
        output=multiprocessing.Queue(),
        input=multiprocessing.Queue(),
    )
    yield queues
    # Prevent BrokenPipeError tracebacks on stderr
    # https://bugs.python.org/issue35844
    time.sleep(0.1)
    queues.output.cancel_join_thread()
    queues.input.cancel_join_thread()


@patch('upsies.utils.torrent.create')
def test_torrent_process_creates_torrent(create_mock, queues):
    create_mock.return_value = 'path/to/foo.mkv.torrent'
    _torrent_process(queues.output, queues.input, some='argument', another='one')
    assert queues.output.get() == (MsgType.result, 'path/to/foo.mkv.torrent')
    assert queues.output.empty()
    assert queues.input.empty()
    assert create_mock.call_args_list == [call(
        some='argument',
        another='one',
        init_callback=Callable(),
        progress_callback=Callable(),
    )]


@patch('upsies.utils.torrent.create')
def test_torrent_process_catches_TorrentCreateError(create_mock, queues):
    create_mock.side_effect = errors.TorrentCreateError('Argh')
    _torrent_process(queues.output, queues.input, some='argument')
    assert queues.output.get() == (MsgType.error, 'Argh')
    assert queues.output.empty()
    assert queues.input.empty()


def test_torrent_process_initializes_with_file_tree(mocker, queues):
    def create_mock(init_callback, **kwargs):
        init_callback('this is not a file tree')

    mocker.patch('upsies.utils.torrent.create', create_mock)
    _torrent_process(queues.output, queues.input, some='argument')
    assert queues.output.get() == (MsgType.init, 'this is not a file tree')


def test_torrent_process_sends_progress(mocker, queues):
    def create_mock(progress_callback, **kwargs):
        for progress in (10, 50, 100):
            progress_callback(progress)

    mocker.patch('upsies.utils.torrent.create', create_mock)
    _torrent_process(queues.output, queues.input, some='argument')
    assert queues.output.get() == (MsgType.info, 10)
    assert queues.output.get() == (MsgType.info, 50)
    assert queues.output.get() == (MsgType.info, 100)


@pytest.mark.parametrize(
    argnames='callback_name, callback_args, msgs, exp_output',
    argvalues=(
        pytest.param(
            'init_callback',
            [
                'mock file tree 1',
                'mock file tree 2',
                'mock file tree 3',
            ],
            [
                (None, None),
                (MsgType.terminate, 'ignored'),
                (None, None),
            ],
            [
                (MsgType.init, 'mock file tree 1'),
                (MsgType.result, 'cb returned: cancel'),
            ],
            id='init_callback provides file_tree and cancels',
        ),

        pytest.param(
            'progress_callback',
            [
                make_CreateTorrentProgress(percent_done=12),
                make_CreateTorrentProgress(percent_done=75),
                make_CreateTorrentProgress(percent_done=100),
            ],
            [
                (None, None),
                (None, None),
                (None, None),
            ],
            [
                (MsgType.info, make_CreateTorrentProgress(percent_done=12)),
                (MsgType.info, make_CreateTorrentProgress(percent_done=75)),
                (MsgType.info, make_CreateTorrentProgress(percent_done=100)),
                (MsgType.result, 'mock/path/to.torrent'),
            ],
            id='progress_callback gets CreateTorrentProgress until torrent is finished',
        ),

        pytest.param(
            'progress_callback',
            [
                make_CreateTorrentProgress(percent_done=12),
                make_CreateTorrentProgress(percent_done=75),
                make_CreateTorrentProgress(percent_done=100),
            ],
            [
                (None, None),
                (MsgType.terminate, None),
                (None, None),
            ],
            [
                (MsgType.info, make_CreateTorrentProgress(percent_done=12)),
                (MsgType.result, 'cb returned: cancel'),
            ],
            id='progress_callback gets CreateTorrentProgress and cancels torrent creation',
        ),

        pytest.param(
            'progress_callback',
            [
                make_CreateTorrentProgress(percent_done=12),
                make_CreateTorrentProgress(percent_done=75),
                make_CreateTorrentProgress(percent_done=100),
            ],
            [
                (None, None),
                (MsgType.terminate, torrent.SKIP_SEARCHING),
                (None, None),
            ],
            [
                (MsgType.info, make_CreateTorrentProgress(percent_done=12)),
                (MsgType.result, 'cb returned: cancel'),
            ],
            id='progress_callback gets CreateTorrentProgress and tries to skip searching',
        ),

        pytest.param(
            'progress_callback',
            [
                make_FindTorrentProgress(percent_done=12, status=False),
                make_FindTorrentProgress(percent_done=15, status=None),
                make_FindTorrentProgress(percent_done=100, status='never reached'),
            ],
            [
                (None, None),
                (None, None),
                (MsgType.terminate, torrent.SKIP_SEARCHING),
                (None, None),
            ],
            [
                (MsgType.info, make_FindTorrentProgress(percent_done=12, status=False)),
                (MsgType.info, make_FindTorrentProgress(percent_done=15, status=None)),
                (MsgType.result, f'cb returned: {torrent.SKIP_SEARCHING}'),
            ],
            id='progress_callback gets FindTorrentProgress and skips searching',
        ),
    ),
)
def test_torrent_process(callback_name, callback_args, msgs, exp_output, queues, mocker):
    def create_mock(**kwargs):
        callback = kwargs[callback_name]
        for typ, msg in msgs:
            print('### Sending', (typ, msg))
            queues.input.put((typ, msg))
            time.sleep(0.1)

            callback_return_value = callback(callback_args.pop(0))
            print('###', callback_name, 'returned', repr(callback_return_value))
            if callback_return_value:
                return f'cb returned: {callback_return_value}'

        return 'mock/path/to.torrent'

    mocker.patch('upsies.utils.torrent.create', create_mock)
    _torrent_process(queues.output, queues.input, some='argument')

    output = []
    print('### OUTPUT:')
    while True:
        try:
            output.append(queues.output.get(timeout=0.3))
            print('###', output[-1])
        except multiprocessing.queues.Empty:
            break

    assert output == exp_output



import re
from unittest.mock import Mock, PropertyMock, call

import pytest

from upsies import errors
from upsies.trackers import ptp


@pytest.fixture
def make_tracker():
    def make_tracker(**kwargs):
        options = {
            'username': 'bunny',
            'password': 'hunter2',
            'base_url': 'http://ptp.local',
        }
        options.update(kwargs)
        return ptp.PtpTracker(options=options)

    return make_tracker

@pytest.fixture
def tracker(make_tracker):
    return make_tracker()


class MockResponse(str):
    def __new__(cls, html='<html>nothing</html>', headers={}, json={}):
        self = super().__new__(cls, html)
        self.headers = headers
        self.json = Mock(return_value=json)
        return self


def test_name_attribute():
    assert ptp.PtpTracker.name == 'ptp'


def test_label_attribute():
    assert ptp.PtpTracker.label == 'PTP'


def test_PtpTracker_TrackerConfig_attribute():
    assert ptp.PtpTracker.TrackerConfig is ptp.PtpTrackerConfig


def test_PtpTracker_TrackerJobs_attribute():
    assert ptp.PtpTracker.TrackerJobs is ptp.PtpTrackerJobs


def test_PtpTracker_base_url_property(tracker, mocker):
    mocker.patch.object(type(tracker), '_base_url', PropertyMock(return_value='http://foo.local'))
    assert tracker._base_url == 'http://foo.local'


def test_PtpTracker_ajax_url_property(tracker, mocker):
    mocker.patch.object(type(tracker), '_base_url', PropertyMock(return_value='http://foo.local'))
    assert tracker._ajax_url == 'http://foo.local/ajax.php'


def test_PtpTracker_logout_url_property(tracker, mocker):
    mocker.patch.object(type(tracker), '_base_url', PropertyMock(return_value='http://foo.local'))
    assert tracker._logout_url == 'http://foo.local/logout.php'


def test_PtpTracker_upload_url_property(tracker, mocker):
    mocker.patch.object(type(tracker), '_base_url', PropertyMock(return_value='http://foo.local'))
    assert tracker._upload_url == 'http://foo.local/upload.php'


def test_PtpTracker_torrents_url_property(tracker, mocker):
    mocker.patch.object(type(tracker), '_base_url', PropertyMock(return_value='http://foo.local'))
    assert tracker._torrents_url == 'http://foo.local/torrents.php'


@pytest.mark.parametrize(
    argnames='announce_url, exp_result',
    argvalues=(
        ('', errors.AnnounceUrlNotSetError(tracker='mock tracker instance')),
        ('http://mock.announce.url', 'http://mock.announce.url'),
    ),
)
def test_PtpTracker_announce_url_property(announce_url, exp_result, tracker, mocker):
    mocker.patch.dict(tracker.options, {'announce_url': announce_url})

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            tracker._announce_url
    else:
        assert tracker._announce_url == exp_result


@pytest.mark.parametrize(
    argnames='announce_url, exp_result',
    argvalues=(
        ('http://mock.url:1234/d34db33f/announce', 'd34db33f'),
        ('http://mock.url:1234/announce', RuntimeError('Failed to find passkey in announce URL: http://mock.url:1234/announce')),
    ),
)
def test_PtpTracker_passkey_property(announce_url, exp_result, tracker, mocker):
    mocker.patch.object(type(tracker), '_announce_url', PropertyMock(return_value=announce_url))

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            tracker._passkey
    else:
        assert tracker._passkey == exp_result


@pytest.fixture
def tracker_for_request(tracker, mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch('upsies.trackers.mtv.tracker.utils.http.get'), 'get')
    mocks.attach_mock(mocker.patch('upsies.trackers.mtv.tracker.utils.http.post'), 'post')
    mocks.attach_mock(mocker.patch.object(tracker, '_maybe_raise_error'), '_maybe_raise_error')
    return tracker, mocks

@pytest.mark.parametrize('method', ('GET', 'get', 'POST', 'post'))
@pytest.mark.asyncio
async def test_PtpTracker_request_catches_RequestError_from_method(method, tracker_for_request):
    tracker, mocks = tracker_for_request
    get_exception = errors.RequestError('Your GET request sucks')
    post_exception = errors.RequestError('Your POST request sucks')
    mocks.get.side_effect = get_exception
    mocks.post.side_effect = post_exception

    args = ('foo', 'bar')
    kwargs = {'baz': 123}

    await tracker._request(method, *args, **kwargs)

    assert mocks.mock_calls == [
        getattr(call, method.lower())(
            *args,
            user_agent=True,
            follow_redirects=False,
            **kwargs,
        ),
        call._maybe_raise_error(
            get_exception if method.lower() == 'get' else post_exception
        ),
    ]

@pytest.mark.parametrize('method', ('GET', 'POST'))
@pytest.mark.parametrize(
    argnames='error_prefix, exception, exp_result',
    argvalues=(
        (None, errors.RequestError('Raised error'), errors.RequestError('Raised error')),
        ('', errors.RequestError('Raised error'), errors.RequestError('Raised error')),
        ('YO', errors.RequestError('Raised error'), errors.RequestError('YO: Raised error')),
        ('YO', None, '<method return value>'),
    ),
)
@pytest.mark.asyncio
async def test_PtpTracker_request_catches_RequestError_from_maybe_raise_error(method, error_prefix, exception,
                                                                              exp_result,
                                                                              tracker_for_request):
    tracker, mocks = tracker_for_request
    mocks._maybe_raise_error.side_effect = exception

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            await tracker._request(method, error_prefix=error_prefix)
    else:
        return_value = await tracker._request(method, error_prefix=error_prefix)
        assert return_value is getattr(mocks, method.lower()).return_value

    assert mocks.mock_calls == [
        getattr(call, method.lower())(
            user_agent=True,
            follow_redirects=False,
        ),
        call._maybe_raise_error(getattr(mocks, method.lower()).return_value),
    ]


@pytest.mark.parametrize(
    argnames='response_or_request_error, exp_exception',
    argvalues=(
        ('normal response', None),
        (errors.RequestError('Raised error'), errors.RequestError('Raised error')),
    ),
    ids=lambda v: repr(v),
)
def test_PtpTracker_maybe_raise_error(response_or_request_error, exp_exception, tracker, mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(tracker, '_maybe_raise_error_from_json'), '_maybe_raise_error_from_json')
    mocks.attach_mock(mocker.patch.object(tracker, '_maybe_raise_error_from_html'), '_maybe_raise_error_from_html')

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            tracker._maybe_raise_error(response_or_request_error)
    else:
        return_value = tracker._maybe_raise_error(response_or_request_error)
        assert return_value is None

    assert mocks.mock_calls == [
        call._maybe_raise_error_from_json(response_or_request_error),
        call._maybe_raise_error_from_html(response_or_request_error),
    ]


@pytest.mark.parametrize(
    argnames='json, exp_exception',
    argvalues=(
        (Mock(side_effect=errors.RequestError('Malformed JSON')), None),
        (Mock(return_value={}), None),
        (Mock(return_value={'Result': 'Success'}), None),
        (Mock(return_value={'Result': 'Error'}), None),
        (Mock(return_value={'Result': 'Error', 'Message': ''}), None),
        (
            Mock(return_value={'Result': 'Error', 'Message': 'Oh no!'}),
            errors.RequestError('Oh no!'),
        ),
        (
            Mock(return_value={'Result': 'Error', 'Message': 'Oh no!\n<br />\n<br>This is not good.<br>\n<br/>'}),
            errors.RequestError('Oh no!\nThis is not good.'),
        ),
    ),
    ids=lambda v: repr(v),
)
def test_PtpTracker_maybe_raise_error_from_json(json, exp_exception, tracker):
    response_or_request_error = Mock(json=json)

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            tracker._maybe_raise_error_from_json(response_or_request_error)
    else:
        return_value = tracker._maybe_raise_error_from_json(response_or_request_error)
        assert return_value is None


@pytest.mark.parametrize('wrapper_class', (str, errors.RequestError))
@pytest.mark.parametrize(
    argnames='text, exp_exception',
    argvalues=(
        ('This is not HTML', None),
        ('<html>\nThis is HTML\n</html>', None),
        ('<html>\n<div id="content"><div>This is not an error</div></div>\n</html>', None),
        ('<html>\n<div id="content"><div class="page__title">Good News!</div></div>\n</html>', None),
        ('<html>\n<div id="content"><div class="page__title">Error</div></div>\n</html>', None),
        (
            (
                '<html>\n<div id="content">'
                '<div class="page__title">Error</div>'
                '<div>Everything is great</div>'
                '</div>\n</html>'
            ),
            None,
        ),
        (
            (
                '<html>\n<div id="content">'
                '<div class="page__title">Error</div>'
                '<div class="panel__body"></div>'
                '</div>\n</html>'
            ),
            None,
        ),
        (
            (
                '<html>\n<div id="content">'
                '<div class="page__title">Error</div>'
                '<div class="panel__body">  Something went wrong<br/>  Badly!  </div>'
                '</div>\n</html>'
            ),
            errors.RequestError('Something went wrong  Badly!'),
        ),
    ),
    ids=lambda v: repr(v),
)
def test_PtpTracker_maybe_raise_error_from_html(text, wrapper_class, exp_exception, tracker):
    response_or_request_error = wrapper_class(text)

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            tracker._maybe_raise_error_from_html(response_or_request_error)
    else:
        return_value = tracker._maybe_raise_error_from_html(response_or_request_error)
        assert return_value is None


def test_PtpTracker_get_anti_csrf_token(tracker):
    response = Mock(json=Mock(return_value={'AntiCsrfToken': 'd34db33f'}))
    return_value = tracker._get_anti_csrf_token(response)
    assert return_value == 'd34db33f'


@pytest.mark.parametrize(
    argnames='response, exp_result',
    argvalues=(
        (
            '<html><a href="http://host/logout.php?auth=d34db33f&foo=bar">logout</a></html>',
            'd34db33f'
        ),
        (
            '<html><a href="http://host/logout.php?foo=bar&auth=d34db33f">logout</a></html>',
            'd34db33f'
        ),
        (
            '<html><a href="http://host/foo.php?foo=bar&auth=d34db33f">logout</a></html>',
            RuntimeError('Could not find auth'),
        ),
        (
            '<html><a>logout</a></html>',
            RuntimeError('Could not find auth'),
        ),
        (
            '<html></html>',
            RuntimeError('Could not find auth'),
        ),
        (
            errors.RequestError('Connection refused'),
            errors.RequestError('Connection refused'),
        ),
    ),
)
@pytest.mark.asyncio
async def test_PtpTracker_get_auth(response, exp_result, tracker, mocker):
    mocker.patch.object(type(tracker), 'is_logged_in', PropertyMock(return_value=True))
    mocker.patch.object(type(tracker), '_base_url', PropertyMock(return_value='http://host/'))

    mocks = Mock()
    if isinstance(response, Exception):
        mocks.attach_mock(
            mocker.patch.object(tracker, '_request', side_effect=response),
            '_request',
        )
    else:
        mocks.attach_mock(
            mocker.patch.object(tracker, '_request', return_value=response),
            '_request',
        )
    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            await tracker._get_auth()
    else:
        return_value = await tracker._get_auth()
        assert return_value == exp_result

    assert mocks.mock_calls == [
        call._request('GET', tracker._base_url),
    ]


@pytest.mark.parametrize(
    argnames='options, exp_result',
    argvalues=(
        pytest.param(
            {'password': 'hunter2'},
            errors.RequestError('Login failed: No username configured'),
            id='No username configured',
        ),
        pytest.param(
            {'username': 'AzureDiamond'},
            errors.RequestError('Login failed: No password configured'),
            id='No password configured',
        ),
        pytest.param(
            {'username': 'AzureDiamond', 'password': 'hunter2'},
            None,
            id='Login succeeds',
        ),
    ),
)
@pytest.mark.asyncio
async def test_PtpTracker_login(options, exp_result, tracker, mocker):
    mocker.patch.object(type(tracker), 'options', PropertyMock(return_value=options))
    mocker.patch.object(type(tracker), '_ajax_url', PropertyMock(return_value='http://host/ajax.php'))
    mocker.patch.object(type(tracker), '_passkey', PropertyMock(return_value='mypasskey'))

    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(tracker, '_request'), '_request')
    mocks.attach_mock(
        mocker.patch.object(tracker, '_get_anti_csrf_token', return_value='MOCK_ANTI_CSRF_TOKEN'),
        '_get_anti_csrf_token',
    )

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            await tracker._login()

        assert mocks.mock_calls == []
        assert not hasattr(tracker, '_anti_csrf_token')

    else:
        return_value = await tracker._login()
        assert return_value is None

        assert mocks.mock_calls == [
            call._request(
                method='POST',
                url=f'{tracker._ajax_url}?action=login',
                data={
                    'username': options['username'],
                    'password': options['password'],
                    'passkey': tracker._passkey,
                },
                error_prefix='Login failed',
                debug_file='http_dumps.ptp/ptp_login',
            ),
            call._get_anti_csrf_token(mocks._request.return_value),
        ]
        assert tracker._anti_csrf_token == 'MOCK_ANTI_CSRF_TOKEN'


@pytest.mark.parametrize('anti_csrf_token', (None, 'MOCK_ANTI_CSRF_TOKEN'))
@pytest.mark.parametrize(
    argnames='response',
    argvalues=(
        pytest.param(
            'ignored mock response',
            id='Logout succeeds',
        ),
        pytest.param(
            errors.RequestError('YOU CANNOT LEAVE'),
            id='Logout raises',
        ),
    ),
)
@pytest.mark.asyncio
async def test_PtpTracker_logout(response, anti_csrf_token, tracker, mocker):
    mocks = Mock()
    mocker.patch.object(type(tracker), 'name', PropertyMock(return_value='mytracker'))
    mocker.patch.object(type(tracker), '_logout_url', PropertyMock(return_value='http://host/logout.php'))
    if anti_csrf_token is not None:
        tracker._anti_csrf_token = anti_csrf_token
    mocks.attach_mock(
        mocker.patch.object(tracker, '_get_auth', return_value='mock auth'),
        '_get_auth',
    )
    if isinstance(response, Exception):
        mocks.attach_mock(
            mocker.patch.object(tracker, '_request', side_effect=response),
            '_request',
        )
    else:
        mocks.attach_mock(
            mocker.patch.object(tracker, '_request', return_value=response),
            '_request',
        )

    if isinstance(response, Exception):
        with pytest.raises(type(response), match=rf'^{re.escape(str(response))}$'):
            await tracker._logout()
    else:
        return_value = await tracker._logout()
        assert return_value is None

    assert mocks.mock_calls == [
        call._get_auth(),
        call._request(
            method='GET',
            url=tracker._logout_url,
            params={'auth': mocks._get_auth.return_value},
            error_prefix='Logout failed',
            debug_file='http_dumps.ptp/ptp_logout',
        ),
    ]
    assert not hasattr(tracker, '_anti_csrf_token')


@pytest.mark.asyncio
async def test_PtpTracker_get_announce_url(tracker, mocker):
    mocker.patch.object(type(tracker), '_announce_url', PropertyMock(return_value='http://mock.announce.url'))
    announce_url = await tracker.get_announce_url()
    assert announce_url == 'http://mock.announce.url'


@pytest.mark.asyncio
async def test_PtpTracker_upload(tracker, mocker):
    mocks = Mock()
    mocks.attach_mock(
        mocker.patch('upsies.utils.http.post'),
        'post',
    )
    mocks.attach_mock(
        mocker.patch.object(tracker, '_handle_upload_response', return_value='mock response'),
        '_handle_upload_response',
    )
    mocker.patch.object(tracker, '_anti_csrf_token', 'MOCK_ANTI_CSRF_TOKEN', create=True)
    mocker.patch.object(type(tracker), '_upload_url', PropertyMock(return_value='http://host/upload.php'))

    tracker_jobs = Mock(
        post_data={'mock': 'post', 'data': 'for you'},
        torrent_filepath='path/to/content.torrent',
    )

    return_value = await tracker.upload(tracker_jobs)
    assert return_value == mocks._handle_upload_response.return_value

    assert mocks.mock_calls == [
        call.post(
            url=tracker._upload_url,
            cache=False,
            user_agent=True,
            data={'mock': 'post', 'data': 'for you', 'AntiCsrfToken': 'MOCK_ANTI_CSRF_TOKEN'},
            files={
                'file_input': {
                    'file': tracker_jobs.torrent_filepath,
                    'mimetype': 'application/x-bittorrent',
                },
            },
            follow_redirects=False,
            debug_file='http_dumps.ptp/ptp_upload',
        ),
        call._handle_upload_response(mocks.post.return_value),
    ]


@pytest.mark.parametrize(
    argnames='response, exp_result, exp_calls',
    argvalues=(
        pytest.param(
            MockResponse(headers={'Location': 'torrents.php?id=123'}),
            '{base_url}/torrents.php?id=123',
            [],
            id='Redirect to torrent page URL',
        ),
        pytest.param(
            MockResponse(headers={'Location': 'foo.php?id=123'}),
            errors.RequestError('Failed to interpret response (see ptp_upload_failed.html)'),
            [
                call.html_dump(str(MockResponse()), 'ptp_upload_failed.html'),
            ],
            id='Redirect to other URL',
        ),
        pytest.param(
            MockResponse(),
            errors.RequestError('Failed to interpret response (see ptp_upload_failed.html)'),
            [
                call.html_dump(str(MockResponse()), 'ptp_upload_failed.html'),
            ],
            id='No redirect',
        ),
        pytest.param(
            MockResponse(html='<html><div class="alert">  Your upload is <b>no good</b>!\n</html>'),
            errors.RequestError('Upload failed: Your upload is no good!'),
            [],
            id='Error message is found',
        ),
    ),
)
def test_PtpTracker_handle_upload_response(response, exp_result, exp_calls, tracker, mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(tracker, 'error'), 'error')
    mocks.attach_mock(mocker.patch('upsies.utils.html.dump'), 'html_dump')

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            tracker._handle_upload_response(response)
    else:
        return_value = tracker._handle_upload_response(response)
        assert return_value == exp_result.format(base_url=tracker._base_url)

    assert mocks.mock_calls == exp_calls


@pytest.mark.parametrize(
    argnames='imdb_id_digits, exp_return_value',
    argvalues=(
        ('1234567890', '1234567890'),
        ('123456789', '123456789'),
        ('12345678', '12345678'),
        ('1234567', '1234567'),
        ('123456', '0123456'),
        ('12345', '0012345'),
        ('1234', '0001234'),
        ('123', '0000123'),
        ('12', '0000012'),
        ('1', '0000001'),
        ('0', '0'),
        (1234567890, '1234567890'),
        (123456789, '123456789'),
        (12345678, '12345678'),
        (1234567, '1234567'),
        (123456, '0123456'),
        (12345, '0012345'),
        (1234, '0001234'),
        (123, '0000123'),
        (12, '0000012'),
        (1, '0000001'),
        (0, '0'),
        ('', '0'),
        (None, '0'),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.parametrize('prefix', ('tt', ''))
@pytest.mark.asyncio
async def test_PtpTracker_normalize_imdb_id(prefix, imdb_id_digits, exp_return_value, tracker):
    if isinstance(imdb_id_digits, str):
        imdb_id = prefix + imdb_id_digits
    else:
        imdb_id = imdb_id_digits

    return_value = tracker.normalize_imdb_id(imdb_id)
    assert return_value == exp_return_value


@pytest.mark.parametrize(
    argnames='imdb_id, exceptions, response, exp_return_value, exp_mock_calls',
    argvalues=(
        pytest.param(
            'tt0123',
            {},
            MockResponse(headers={'location': 'foo.php?id=123456'}),
            '123456',
            [
                call.login(),
                call.normalize_imdb_id('tt0123'),
                call._request,
            ],
            id='Group ID is found',
        ),
        pytest.param(
            'tt0123',
            {},
            MockResponse(headers={'location': 'foo.php?this=that'}),
            None,
            [
                call.login(),
                call.normalize_imdb_id('tt0123'),
                call._request,
            ],
            id='No group ID in location header',
        ),
        pytest.param(
            'tt0123',
            {},
            MockResponse(headers={'foo': 'bar'}),
            None,
            [
                call.login(),
                call.normalize_imdb_id('tt0123'),
                call._request,
            ],
            id='No location header in response',
        ),
        pytest.param(
            'tt0123',
            {
                'login': errors.RequestError('Wrong password'),
            },
            None,
            None,
            [
                call.login(),
                call.error(errors.RequestError('Wrong password')),
            ],
            id='Login failed',
        ),
        pytest.param(
            'tt0123',
            {
                '_request': errors.RequestError('No response'),
            },
            None,
            None,
            [
                call.login(),
                call.normalize_imdb_id('tt0123'),
                call._request,
                call.error(errors.RequestError('No response')),
            ],
            id='Request failed',
        ),
    ),
)
@pytest.mark.asyncio
async def test_PtpTracker_get_ptp_group_id_by_imdb_id(imdb_id, exceptions, response,
                                                      exp_return_value, exp_mock_calls,
                                                      tracker, mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(tracker, 'login'), 'login')
    mocks.attach_mock(mocker.patch.object(tracker, '_request', return_value=response), '_request')
    mocker.patch.object(type(tracker), '_torrents_url', PropertyMock(return_value='http://host/torrents.php'))
    mocks.attach_mock(mocker.patch.object(tracker, 'normalize_imdb_id'), 'normalize_imdb_id')
    mocks.attach_mock(mocker.patch.object(tracker, 'error'), 'error')

    for method, exception in exceptions.items():
        getattr(mocks, method).side_effect = exception

    return_value = await tracker.get_ptp_group_id_by_imdb_id(imdb_id)
    assert return_value == exp_return_value

    # Add call._request() arguments
    if call._request in exp_mock_calls:
        exp_mock_calls[exp_mock_calls.index(call._request)] = call._request(
            method='GET',
            url=tracker._torrents_url,
            params={
                'imdb': tracker.normalize_imdb_id.return_value,
                'json': '1',
            },
            cache=True,
        )
    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize(
    argnames='imdb_id, key, exceptions, response, exp_return_value, exp_mock_calls',
    argvalues=(
        pytest.param(
            'tt0123', None,
            {},
            MockResponse(json=[{'all': 'metadata'}]),
            {'all': 'metadata'},
            [
                call.login(),
                call.normalize_imdb_id('tt0123'),
                call._request,
            ],
            id='Get all metadata',
        ),
        pytest.param(
            'tt0123', 'bar',
            {},
            MockResponse(json=[{'foo': 1, 'bar': 2, 'baz': 3}]),
            2,
            [
                call.login(),
                call.normalize_imdb_id('tt0123'),
                call._request,
            ],
            id='Get single metadata field',
        ),
        pytest.param(
            'tt0123', 'bar',
            {
                'login': errors.RequestError('Wrong password'),
            },
            None,
            None,
            [
                call.login(),
                call.error(errors.RequestError('Wrong password')),
            ],
            id='Login failed',
        ),
        pytest.param(
            'tt0123', 'bar',
            {
                '_request': errors.RequestError('No response'),
            },
            None,
            None,
            [
                call.login(),
                call.normalize_imdb_id('tt0123'),
                call._request,
                call.error(errors.RequestError('No response')),
            ],
            id='Request failed',
        ),
    ),
)
@pytest.mark.asyncio
async def test_PtpTracker_get_ptp_metadata(imdb_id, key, exceptions, response,
                                           exp_return_value, exp_mock_calls,
                                           tracker, mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(tracker, 'login'), 'login')
    mocks.attach_mock(mocker.patch.object(tracker, '_request', return_value=response), '_request')
    mocker.patch.object(type(tracker), '_ajax_url', PropertyMock(return_value='http://host/ajax.php'))
    mocks.attach_mock(mocker.patch.object(tracker, 'normalize_imdb_id'), 'normalize_imdb_id')
    mocks.attach_mock(mocker.patch.object(tracker, 'error'), 'error')

    for method, exception in exceptions.items():
        getattr(mocks, method).side_effect = exception

    return_value = await tracker.get_ptp_metadata(imdb_id, key=key)
    assert return_value == exp_return_value

    # Add call._request() arguments
    if call._request in exp_mock_calls:
        exp_mock_calls[exp_mock_calls.index(call._request)] = call._request(
            method='GET',
            url='http://host/ajax.php',
            params={
                'action': 'torrent_info',
                'imdb': mocks.normalize_imdb_id.return_value,
            },
            cache=True,
        )
    assert mocks.mock_calls == exp_mock_calls


@pytest.mark.parametrize(
    argnames='bytes, exp_piece_size',
    argvalues=(
        # Numbers are copied from PTP piece size chart
        (1 * 1024 * 1024, 2**15),
        (6 * 1024 * 1024, 2**15),
        (10 * 1024 * 1024, 2**15),
        (12 * 1024 * 1024, 2**15),
        (15 * 1024 * 1024, 2**15),
        (18 * 1024 * 1024, 2**15),
        (22 * 1024 * 1024, 2**15),
        (27 * 1024 * 1024, 2**15),

        (33 * 1024 * 1024, 2**16),
        (40 * 1024 * 1024, 2**16),
        (48 * 1024 * 1024, 2**16),
        (58 * 1024 * 1024, 2**16),

        (70 * 1024 * 1024, 2**17),
        (84 * 1024 * 1024, 2**17),
        (101 * 1024 * 1024, 2**17),
        (122 * 1024 * 1024, 2**17),

        (147 * 1024 * 1024, 2**18),
        (177 * 1024 * 1024, 2**18),
        (213 * 1024 * 1024, 2**18),
        (256 * 1024 * 1024, 2**18),

        (308 * 1024 * 1024, 2**19),
        (370 * 1024 * 1024, 2**19),
        (444 * 1024 * 1024, 2**19),

        (533 * 1024 * 1024, 2**20),
        (640 * 1024 * 1024, 2**20),
        (768 * 1024 * 1024, 2**20),
        (922 * 1024 * 1024, 2**20),

        (1.08 * 1024 * 1024 * 1024, 2**21),
        (1.30 * 1024 * 1024 * 1024, 2**21),
        (1.56 * 1024 * 1024 * 1024, 2**21),
        (1.87 * 1024 * 1024 * 1024, 2**21),

        (2.24 * 1024 * 1024 * 1024, 2**22),
        (2.69 * 1024 * 1024 * 1024, 2**22),
        (3.23 * 1024 * 1024 * 1024, 2**22),
        (3.88 * 1024 * 1024 * 1024, 2**22),

        (4.65 * 1024 * 1024 * 1024, 2**23),
        (5.59 * 1024 * 1024 * 1024, 2**23),
        (6.70 * 1024 * 1024 * 1024, 2**23),
        (8.04 * 1024 * 1024 * 1024, 2**23),

        (9.65 * 1024 * 1024 * 1024, 2**24),
        (11.58 * 1024 * 1024 * 1024, 2**24),
        (13.90 * 1024 * 1024 * 1024, 2**24),
        (16.68 * 1024 * 1024 * 1024, 2**24),
        (20.02 * 1024 * 1024 * 1024, 2**24),
        (24.02 * 1024 * 1024 * 1024, 2**24),
        (28.83 * 1024 * 1024 * 1024, 2**24),
        (34.59 * 1024 * 1024 * 1024, 2**24),
        (41.51 * 1024 * 1024 * 1024, 2**24),
        (49.81 * 1024 * 1024 * 1024, 2**24),
        (59.78 * 1024 * 1024 * 1024, 2**24),
        (71.73 * 1024 * 1024 * 1024, 2**24),
        (86.08 * 1024 * 1024 * 1024, 2**24),
        (103.30 * 1024 * 1024 * 1024, 2**24),
        (123.96 * 1024 * 1024 * 1024, 2**24),
        (148.75 * 1024 * 1024 * 1024, 2**24),
        (178.50 * 1024 * 1024 * 1024, 2**24),
        (214.20 * 1024 * 1024 * 1024, 2**24),
        (257.04 * 1024 * 1024 * 1024, 2**24),
        (308.45 * 1024 * 1024 * 1024, 2**24),
        (370.14 * 1024 * 1024 * 1024, 2**24),
        (444.16 * 1024 * 1024 * 1024, 2**24),
        (533.00 * 1024 * 1024 * 1024, 2**24),
        (639.60 * 1024 * 1024 * 1024, 2**24),
        (767.52 * 1024 * 1024 * 1024, 2**24),
        (921.02 * 1024 * 1024 * 1024, 2**24),
    ),
)
def test_PtpTracker_calculate_piece_size(bytes, exp_piece_size):
    piece_size = ptp.PtpTracker.calculate_piece_size(bytes)
    assert piece_size == exp_piece_size


@pytest.mark.parametrize(
    argnames='bytes, exp_result',
    argvalues=(
        (-1, ValueError('Unexpected size: -1')),
        (0, (2**15, 2**16)),
        (1, (2**15, 2**16)),
        (147 * 1024 * 1024 - 1, (2**15, 2**16)),
        (147 * 1024 * 1024, (2**16, 2**19)),
        (256 * 1024 * 1024, (2**17, 2**20)),
        (533 * 1024 * 1024, (2**18, 2**21)),
        (1.08 * 1024 * 1024 * 1024 - 1, (2**18, 2**21)),
        (1.08 * 1024 * 1024 * 1024, (2**19, 2**22)),
        (2.24 * 1024 * 1024 * 1024 - 1, (2**19, 2**22)),
        (2.24 * 1024 * 1024 * 1024, (2**20, 2**23)),
        (4.65 * 1024 * 1024 * 1024 - 1, (2**20, 2**23)),
        (4.65 * 1024 * 1024 * 1024, (2**21, 2**24)),
        (8.04 * 1024 * 1024 * 1024 - 1, (2**21, 2**24)),
        (8.04 * 1024 * 1024 * 1024, (2**22, 2**24)),
        (148.75 * 1024 * 1024 * 1024 - 1, (2**22, 2**24)),
        (148.75 * 1024 * 1024 * 1024, (2**23, 2**24)),
        (257.04 * 1024 * 1024 * 1024 - 1, (2**23, 2**24)),
        (257.04 * 1024 * 1024 * 1024, (2**24, 2**24)),
        (float('inf'), (2**24, 2**24)),
    ),
    ids=lambda v: repr(v),
)
def test_PtpTracker_calculate_piece_size_min_max(bytes, exp_result):
    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            ptp.PtpTracker.calculate_piece_size_min_max(bytes)
    else:
        piece_size_min, piece_size_max = ptp.PtpTracker.calculate_piece_size_min_max(bytes)
        assert (piece_size_min, piece_size_max) == exp_result



import re
from unittest.mock import AsyncMock, MagicMock, Mock, PropertyMock, call

import pytest

from upsies import __homepage__, __project_name__, errors, utils
from upsies.trackers import mtv


@pytest.fixture
def tracker():
    tracker = Mock()
    tracker.name = 'mtv'
    tracker.options = {
        'exclude': [],
    }
    return tracker


@pytest.fixture
def imghost():
    class MockImageHost(utils.imghosts.base.ImageHostBase):
        name = 'mock image host'
        default_config = {}
        _upload_image = AsyncMock()

    return MockImageHost()


@pytest.fixture
def mtv_tracker_jobs(imghost, tracker, tmp_path, mocker):
    content_path = tmp_path / 'Foo 2000 1080p BluRay x264-ASDF'

    mtv_tracker_jobs = mtv.MtvTrackerJobs(
        content_path=str(content_path),
        tracker=tracker,
        image_host=imghost,
        btclient=Mock(),
        torrent_destination=str(tmp_path / 'destination'),
        common_job_args={
            'home_directory': str(tmp_path / 'home_directory'),
            'ignore_cache': True,
        },
        options=None,
    )

    return mtv_tracker_jobs


@pytest.fixture
def mock_job_attributes(mocker):
    def mock_job_attributes(mtv_tracker_jobs):
        job_attrs = (
            # Background jobs
            'create_torrent_job',
            'mediainfo_job',
            'screenshots_job',
            'upload_screenshots_job',

            # Interactive jobs
            'category_job',
            'imdb_job',
            'scene_check_job',
            'title_job',
            'description_job',
        )
        for job_attr in job_attrs:
            mocker.patch.object(type(mtv_tracker_jobs), job_attr, PropertyMock(return_value=Mock(attr=job_attr)))

    return mock_job_attributes


def test_MtvTrackerJobs_jobs_before_upload(mock_job_attributes, mtv_tracker_jobs):
    mock_job_attributes(mtv_tracker_jobs)

    print(mtv_tracker_jobs.jobs_before_upload)
    assert tuple(job.attr for job in mtv_tracker_jobs.jobs_before_upload) == (
        # Background jobs
        'create_torrent_job',
        'mediainfo_job',
        'screenshots_job',
        'upload_screenshots_job',

        # Interactive jobs
        'category_job',
        'imdb_job',
        'scene_check_job',
        'title_job',
        'description_job',
    )


def test_isolated_jobs__only_description(mtv_tracker_jobs, mock_job_attributes, mocker):
    mock_job_attributes(mtv_tracker_jobs)
    mocker.patch.object(type(mtv_tracker_jobs), 'options', PropertyMock(return_value={'only_description': True}))
    mocker.patch.object(mtv_tracker_jobs, 'get_job_and_dependencies')
    assert mtv_tracker_jobs.isolated_jobs is mtv_tracker_jobs.get_job_and_dependencies.return_value
    assert mtv_tracker_jobs.get_job_and_dependencies.call_args_list == [
        call(mtv_tracker_jobs.description_job, mtv_tracker_jobs.screenshots_job)
    ]

def test_isolated_jobs__only_title(mtv_tracker_jobs, mock_job_attributes, mocker):
    mock_job_attributes(mtv_tracker_jobs)
    mocker.patch.object(type(mtv_tracker_jobs), 'options', PropertyMock(return_value={'only_title': True}))
    mocker.patch.object(mtv_tracker_jobs, 'get_job_and_dependencies')
    assert mtv_tracker_jobs.isolated_jobs is mtv_tracker_jobs.get_job_and_dependencies.return_value
    assert mtv_tracker_jobs.get_job_and_dependencies.call_args_list == [
        call(mtv_tracker_jobs.title_job)
    ]

def test_isolated_jobs__no_isolated_jobs(mtv_tracker_jobs, mock_job_attributes, mocker):
    mock_job_attributes(mtv_tracker_jobs)
    mocker.patch.object(type(mtv_tracker_jobs), 'options', PropertyMock(return_value={}))
    mocker.patch.object(mtv_tracker_jobs, 'get_job_and_dependencies')
    assert mtv_tracker_jobs.isolated_jobs == ()
    assert mtv_tracker_jobs.get_job_and_dependencies.call_args_list == []


def test_MtvTrackerJobs_category_job(mtv_tracker_jobs, mocker):
    ChoiceJob_mock = mocker.patch('upsies.jobs.dialog.ChoiceJob')
    mocker.patch.object(mtv_tracker_jobs, 'get_job_name')
    mocker.patch.object(mtv_tracker_jobs, 'make_precondition')
    mocker.patch.object(mtv_tracker_jobs, 'autodetect_category')
    mocker.patch.object(type(mtv_tracker_jobs), '_categories', PropertyMock(return_value=[
        {'label': 'movie', 'value': '1'},
        {'label': 'series', 'value': '2'},
        {'label': 'vegetable', 'value': '3'},
    ]))
    mocker.patch.object(mtv_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert mtv_tracker_jobs.category_job is ChoiceJob_mock.return_value
    assert ChoiceJob_mock.call_args_list == [call(
        name=mtv_tracker_jobs.get_job_name.return_value,
        label='Category',
        precondition=mtv_tracker_jobs.make_precondition.return_value,
        autodetect=mtv_tracker_jobs.autodetect_category,
        choices=[
            ('movie', '1'),
            ('series', '2'),
            ('vegetable', '3'),
        ],
        callbacks={
            'finished': mtv_tracker_jobs.update_imdb_query,
        },
        common_job_arg='common job argument',
    )]

    assert mtv_tracker_jobs.get_job_name.call_args_list == [call('category')]
    assert mtv_tracker_jobs.make_precondition.call_args_list == [call('category_job')]
    assert mtv_tracker_jobs.common_job_args.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='old_type, new_type, exp_type',
    argvalues=(
        (utils.release.ReleaseType.movie, utils.release.ReleaseType.season, utils.release.ReleaseType.season),
        (utils.release.ReleaseType.movie, utils.release.ReleaseType.episode, utils.release.ReleaseType.episode),
        (utils.release.ReleaseType.movie, utils.release.ReleaseType.unknown, utils.release.ReleaseType.unknown),
        (utils.release.ReleaseType.season, utils.release.ReleaseType.movie, utils.release.ReleaseType.movie),
        (utils.release.ReleaseType.season, utils.release.ReleaseType.episode, utils.release.ReleaseType.episode),
        (utils.release.ReleaseType.season, utils.release.ReleaseType.unknown, utils.release.ReleaseType.unknown),
        (utils.release.ReleaseType.episode, utils.release.ReleaseType.movie, utils.release.ReleaseType.movie),
        (utils.release.ReleaseType.episode, utils.release.ReleaseType.season, utils.release.ReleaseType.season),
        (utils.release.ReleaseType.episode, utils.release.ReleaseType.unknown, utils.release.ReleaseType.unknown),
        (utils.release.ReleaseType.unknown, utils.release.ReleaseType.movie, utils.release.ReleaseType.movie),
        (utils.release.ReleaseType.unknown, utils.release.ReleaseType.season, utils.release.ReleaseType.season),
        (utils.release.ReleaseType.unknown, utils.release.ReleaseType.episode, utils.release.ReleaseType.episode),
        (None, utils.release.ReleaseType.movie, utils.release.ReleaseType.movie),
        (utils.release.ReleaseType.movie, None, utils.release.ReleaseType.movie),
        (utils.release.ReleaseType.season, None, utils.release.ReleaseType.season),
        (utils.release.ReleaseType.episode, None, utils.release.ReleaseType.episode),
    ),
)
def test_update_imdb_query(old_type, new_type, exp_type, mtv_tracker_jobs, mocker):
    mocker.patch.object(type(mtv_tracker_jobs), 'imdb_job', PropertyMock(
        return_value=Mock(
            query=Mock(type=old_type),
        ),
    ))
    mocker.patch.object(type(mtv_tracker_jobs), 'chosen_release_type', PropertyMock(
        return_value=new_type,
    ))
    assert mtv_tracker_jobs.imdb_job.query.type == old_type
    mtv_tracker_jobs.update_imdb_query('ignored category_job')
    assert mtv_tracker_jobs.imdb_job.query.type == exp_type


@pytest.mark.parametrize(
    argnames='release_type, exp_typ',
    argvalues=(
        (utils.release.ReleaseType.movie, 'Movie'),
        (utils.release.ReleaseType.season, 'Season'),
        (utils.release.ReleaseType.episode, 'Episode'),
        ('wat', RuntimeError('Unsupported type: wat')),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.parametrize(
    argnames='height, width, exp_resolution',
    argvalues=(
        (719, 1279, 'SD'),
        (719, 1280, 'HD'),
        (720, 1279, 'HD'),
    ),
    ids=lambda v: repr(v),
)
def test_MtvTrackerJobs_autodetect_category(height, width, exp_resolution, release_type, exp_typ, mtv_tracker_jobs, mocker):
    mocker.patch('upsies.utils.video.height', Mock(return_value=height))
    mocker.patch('upsies.utils.video.width', Mock(return_value=width))
    mocker.patch.object(mtv_tracker_jobs, 'release_name', Mock(type=release_type))

    if isinstance(exp_typ, Exception):
        exp_exception = exp_typ
    else:
        exp_exception = None

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            mtv_tracker_jobs.autodetect_category('mock job')
    else:
        category = mtv_tracker_jobs.autodetect_category('mock job')
        assert category == f'{exp_resolution} {exp_typ}'


@pytest.mark.parametrize(
    argnames='choice',
    argvalues=[c['value'] for c in mtv.MtvTrackerJobs._categories] + [None],
    ids=lambda v: repr(v),
)
@pytest.mark.parametrize(
    argnames='category_job_is_finished',
    argvalues=(True, False),
    ids=lambda v: repr(v),
)
def test_MtvTrackerJobs_chosen_release_type(category_job_is_finished, choice, mtv_tracker_jobs, mocker):
    mocker.patch.object(type(mtv_tracker_jobs), 'category_job', PropertyMock(return_value=Mock(
        is_finished=category_job_is_finished,
        choice=choice,
    )))

    if category_job_is_finished and choice is not None:
        exp_return_value = mtv.MtvTrackerJobs._category_value_type_map[choice]
    else:
        exp_return_value = None

    assert mtv_tracker_jobs.chosen_release_type is exp_return_value


def test_MtvTrackerJobs_release_name_separator(mtv_tracker_jobs, mocker):
    assert mtv_tracker_jobs.release_name_separator == '.'


def test_MtvTrackerJobs_release_name_translation(mtv_tracker_jobs, mocker):
    assert mtv_tracker_jobs.release_name_translation == {
        'edition': {
            re.compile(r"^Director's Cut$"): r'DC',
        },
        'group': {
            re.compile(r'^NOGROUP$'): r'NOGRP',
        },
    }


def test_MtvTrackerJobs_title_job(mtv_tracker_jobs, mocker):
    TextFieldJob_mock = mocker.patch('upsies.jobs.dialog.TextFieldJob')
    mocker.patch.object(mtv_tracker_jobs, 'get_job_name')
    mocker.patch.object(type(mtv_tracker_jobs), 'category_job', PropertyMock())
    mocker.patch.object(type(mtv_tracker_jobs), 'scene_check_job', PropertyMock())
    mocker.patch.object(type(mtv_tracker_jobs), 'imdb_job', PropertyMock())
    mocker.patch.object(mtv_tracker_jobs, 'make_precondition')
    mocker.patch.object(mtv_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert mtv_tracker_jobs.title_job is TextFieldJob_mock.return_value
    assert TextFieldJob_mock.call_args_list == [call(
        name=mtv_tracker_jobs.get_job_name.return_value,
        label='Title',
        prejobs=(
            mtv_tracker_jobs.category_job,
            mtv_tracker_jobs.scene_check_job,
            mtv_tracker_jobs.imdb_job,
        ),
        text=mtv_tracker_jobs.generate_title,
        validator=mtv_tracker_jobs.validate_title,
        precondition=mtv_tracker_jobs.make_precondition.return_value,
        common_job_arg='common job argument',
    )]
    assert mtv_tracker_jobs.get_job_name.call_args_list == [call('title')]
    assert mtv_tracker_jobs.make_precondition.call_args_list == [call('title_job')]
    assert mtv_tracker_jobs.common_job_args.call_args_list == [call()]


@pytest.fixture
def mtv_tracker_jobs_for_generate_title(mtv_tracker_jobs, mocker):
    mocks = Mock()
    mocker.patch.object(type(mtv_tracker_jobs), 'chosen_release_type', PropertyMock())
    mocker.patch.object(type(mtv_tracker_jobs), 'scene_check_job', PropertyMock(return_value=Mock(
        is_finished=True,
        is_scene_release='YOU FORGOT TO SET THIS',
    )))
    mocks.attach_mock(mtv_tracker_jobs.scene_check_job.wait, 'scene_check_job_wait')
    mocks.attach_mock(
        mocker.patch('upsies.utils.predbs.MultiPredbApi.search', AsyncMock()),
        'scene_search',
    )
    mocker.patch.object(type(mtv_tracker_jobs), 'imdb_job', PropertyMock(return_value=Mock(
        is_finished=True,
    )))
    mocker.patch.object(type(mtv_tracker_jobs), 'imdb_id', PropertyMock(return_value='mock imdb_id'))
    mocks.attach_mock(mtv_tracker_jobs.imdb_job.wait, 'imdb_job_wait')
    mocker.patch.object(mtv_tracker_jobs, 'release_name', MagicMock(
        fetch_info=AsyncMock(),
    ))
    mocks.attach_mock(mtv_tracker_jobs.release_name.fetch_info, 'fetch_info')
    return mtv_tracker_jobs, mocks

@pytest.mark.parametrize('job_attribute', ('scene_check_job', 'imdb_job'))
@pytest.mark.asyncio
async def test_MtvTrackerJobs_generate_title__unfinished_depency_job(job_attribute,
                                                                     mtv_tracker_jobs_for_generate_title, mocker):
    mtv_tracker_jobs, mocks = mtv_tracker_jobs_for_generate_title
    getattr(mtv_tracker_jobs, job_attribute).is_finished = False

    with pytest.raises(AssertionError):
        await mtv_tracker_jobs.generate_title()

    assert mocks.mock_calls == []

@pytest.mark.asyncio
async def test_MtvTrackerJobs_generate_title__scene__movie(mtv_tracker_jobs_for_generate_title, mocker):
    mtv_tracker_jobs, mocks = mtv_tracker_jobs_for_generate_title
    mtv_tracker_jobs.scene_check_job.is_scene_release = True
    type(mtv_tracker_jobs).chosen_release_type = PropertyMock(return_value=utils.release.ReleaseType.movie)
    mocks.scene_search.return_value = ['Scene.2000-ASDF', 'Foo.2000-ASDF']
    mtv_tracker_jobs.release_name.__str__.return_value = 'Generated.Title.2012-ARF'

    return_value = await mtv_tracker_jobs.generate_title()
    assert return_value == 'Scene.2000-ASDF'

    assert mocks.mock_calls == [
        call.scene_search(mtv_tracker_jobs.content_path),
    ]

@pytest.mark.asyncio
async def test_MtvTrackerJobs_generate_title__scene__episode(mtv_tracker_jobs_for_generate_title, mocker):
    mtv_tracker_jobs, mocks = mtv_tracker_jobs_for_generate_title
    mtv_tracker_jobs.scene_check_job.is_scene_release = True
    type(mtv_tracker_jobs).chosen_release_type = PropertyMock(return_value=utils.release.ReleaseType.episode)
    mocks.scene_search.return_value = ['Scene.S03E06.2000-ASDF', 'Foo.S04E08.2000-ASDF']
    mtv_tracker_jobs.release_name.__str__.return_value = 'Generated.Title.E02E04-ARF'

    return_value = await mtv_tracker_jobs.generate_title()
    assert return_value == 'Scene.S03E06.2000-ASDF'

    assert mocks.mock_calls == [
        call.scene_search(mtv_tracker_jobs.content_path),
    ]

@pytest.mark.asyncio
async def test_MtvTrackerJobs_generate_title__scene__season(mtv_tracker_jobs_for_generate_title, mocker):
    mtv_tracker_jobs, mocks = mtv_tracker_jobs_for_generate_title
    mtv_tracker_jobs.scene_check_job.is_scene_release = True
    type(mtv_tracker_jobs).chosen_release_type = PropertyMock(return_value=utils.release.ReleaseType.season)
    mocks.scene_search.return_value = ['Scene.S03E06.2000-ASDF', 'Foo.S04E08.2000-ASDF']
    mtv_tracker_jobs.release_name.__str__.return_value = 'Generated.Title.E02E04-ARF'

    return_value = await mtv_tracker_jobs.generate_title()
    assert return_value == 'Generated.Title.E02E04-ARF'

    assert mocks.mock_calls == [
        call.fetch_info(
            webdb=mtv_tracker_jobs.imdb,
            webdb_id=mtv_tracker_jobs.imdb_id,
        ),
    ]

@pytest.mark.asyncio
async def test_MtvTrackerJobs_generate_title__nonscene__movie(mtv_tracker_jobs_for_generate_title, mocker):
    mtv_tracker_jobs, mocks = mtv_tracker_jobs_for_generate_title
    mtv_tracker_jobs.scene_check_job.is_scene_release = False
    type(mtv_tracker_jobs).chosen_release_type = PropertyMock(return_value=utils.release.ReleaseType.movie)
    mocks.scene_search.return_value = ['Scene.2000-ASDF', 'Foo.2000-ASDF']
    mtv_tracker_jobs.release_name.__str__.return_value = 'Generated.Title.2012-ARF'

    return_value = await mtv_tracker_jobs.generate_title()
    assert return_value == 'Generated.Title.2012-ARF'

    assert mocks.mock_calls == [
        call.fetch_info(
            webdb=mtv_tracker_jobs.imdb,
            webdb_id=mtv_tracker_jobs.imdb_id,
        ),
    ]

@pytest.mark.asyncio
async def test_MtvTrackerJobs_generate_title__nonscene__episode(mtv_tracker_jobs_for_generate_title, mocker):
    mtv_tracker_jobs, mocks = mtv_tracker_jobs_for_generate_title
    mtv_tracker_jobs.scene_check_job.is_scene_release = False
    type(mtv_tracker_jobs).chosen_release_type = PropertyMock(return_value=utils.release.ReleaseType.episode)
    mocks.scene_search.return_value = ['Scene.S03E06.2000-ASDF', 'Foo.S04E08.2000-ASDF']
    mtv_tracker_jobs.release_name.__str__.return_value = 'Generated.Title.2012-ARF'

    return_value = await mtv_tracker_jobs.generate_title()
    assert return_value == 'Generated.Title.2012-ARF'

    assert mocks.mock_calls == [
        call.fetch_info(
            webdb=mtv_tracker_jobs.imdb,
            webdb_id=mtv_tracker_jobs.imdb_id,
        ),
    ]

@pytest.mark.asyncio
async def test_MtvTrackerJobs_generate_title__nonscene__season(mtv_tracker_jobs_for_generate_title, mocker):
    mtv_tracker_jobs, mocks = mtv_tracker_jobs_for_generate_title
    mtv_tracker_jobs.scene_check_job.is_scene_release = False
    type(mtv_tracker_jobs).chosen_release_type = PropertyMock(return_value=utils.release.ReleaseType.season)
    mocks.scene_search.return_value = ['Scene.S03.2000-ASDF', 'Foo.S04.2000-ASDF']
    mtv_tracker_jobs.release_name.__str__.return_value = 'Generated.Title.S02-ARF'

    return_value = await mtv_tracker_jobs.generate_title()
    assert return_value == 'Generated.Title.S02-ARF'

    assert mocks.mock_calls == [
        call.fetch_info(
            webdb=mtv_tracker_jobs.imdb,
            webdb_id=mtv_tracker_jobs.imdb_id,
        ),
    ]

@pytest.mark.asyncio
async def test_MtvTrackerJobs_generate_title__getting_title_from_predb_fails(mtv_tracker_jobs_for_generate_title, mocker):
    mtv_tracker_jobs, mocks = mtv_tracker_jobs_for_generate_title
    mtv_tracker_jobs.scene_check_job.is_scene_release = True
    type(mtv_tracker_jobs).chosen_release_type = PropertyMock(return_value=utils.release.ReleaseType.movie)
    mocks.scene_search.side_effect = errors.RequestError('Your request sucks')
    mtv_tracker_jobs.release_name.__str__.return_value = 'Generated.Title-ARF'

    return_value = await mtv_tracker_jobs.generate_title()
    assert return_value == 'Generated.Title-ARF'

    assert mocks.mock_calls == [
        call.scene_search(mtv_tracker_jobs.content_path),
    ]

@pytest.mark.asyncio
async def test_MtvTrackerJobs_generate_title__getting_title_from_webdb_fails(mtv_tracker_jobs_for_generate_title, mocker):
    mtv_tracker_jobs, mocks = mtv_tracker_jobs_for_generate_title
    mtv_tracker_jobs.scene_check_job.is_scene_release = False
    type(mtv_tracker_jobs).chosen_release_type = PropertyMock(return_value='whatever')
    mocks.fetch_info.side_effect = errors.RequestError('Your request sucks')
    mtv_tracker_jobs.release_name.__str__.return_value = 'Generated.Title-ARF'

    return_value = await mtv_tracker_jobs.generate_title()
    assert return_value == 'Generated.Title-ARF'

    assert mocks.mock_calls == [
        call.fetch_info(
            webdb=mtv_tracker_jobs.imdb,
            webdb_id=mtv_tracker_jobs.imdb_id,
        ),
    ]


@pytest.mark.parametrize(
    argnames='title, exp_exception',
    argvalues=(
        ('Valid.Title', None),
        ('', ValueError('Title must not be empty')),
        (' ', ValueError('Title must not be empty')),
        ('    ', ValueError('Title must not be empty')),
        ('Invalid.Title.UNKNOWN_SEASON', ValueError('Replace "UNKNOWN_SEASON" with the correct season')),
        ('Invalid.Title.UNKNOWN_SEASON.720p', ValueError('Replace "UNKNOWN_SEASON" with the correct season')),
        ('Invalid.Title.UNKNOWN_VIDEO_FORMAT', ValueError('Replace "UNKNOWN_VIDEO_FORMAT" with the correct video format')),
        ('Invalid.Title.UNKNOWN_VIDEO_FORMAT.720p', ValueError('Replace "UNKNOWN_VIDEO_FORMAT" with the correct video format')),
    ),
)
def test_MtvTrackerJobs_validate_title(title, exp_exception, mtv_tracker_jobs):
    if isinstance(exp_exception, Exception):
        with pytest.raises(type(exp_exception), match=rf'{re.escape(str(exp_exception))}$'):
            mtv_tracker_jobs.validate_title(title)
    else:
        assert mtv_tracker_jobs.validate_title(title) is None

    def validate_title(self, text):
        if not text.strip():
            raise ValueError('Title must not be empty')

        match = re.search(rf'{utils.release.DELIM}(UNKNOWN_([A-Z_]+)){utils.release.DELIM}', text)
        if match:
            placeholder = match.group(1)
            name = match.group(2)
            raise ValueError(f'Replace "{placeholder}" with the correct {name.lower()}')


def test_MtvTrackerJobs_image_host_config(mtv_tracker_jobs, mocker):
    assert mtv_tracker_jobs.image_host_config == {
        'common': {'thumb_width': 300},
    }


def test_MtvTrackerJobs_description_job(mtv_tracker_jobs, mocker):
    TextFieldJob_mock = mocker.patch('upsies.jobs.dialog.TextFieldJob')
    mocker.patch.object(mtv_tracker_jobs, 'get_job_name')
    mocker.patch.object(type(mtv_tracker_jobs), 'mediainfo_job', PropertyMock())
    mocker.patch.object(type(mtv_tracker_jobs), 'upload_screenshots_job', PropertyMock())
    mocker.patch.object(mtv_tracker_jobs, 'make_precondition')
    mocker.patch.object(mtv_tracker_jobs, 'common_job_args', return_value={'common_job_arg': 'common job argument'})

    assert mtv_tracker_jobs.description_job is TextFieldJob_mock.return_value
    assert TextFieldJob_mock.call_args_list == [call(
        name=mtv_tracker_jobs.get_job_name.return_value,
        label='Description',
        prejobs=(
            mtv_tracker_jobs.mediainfo_job,
            mtv_tracker_jobs.upload_screenshots_job,
        ),
        text=mtv_tracker_jobs.generate_description,
        precondition=mtv_tracker_jobs.make_precondition.return_value,
        finish_on_success=True,
        read_only=True,
        hidden=True,
        common_job_arg='common job argument',
    )]
    assert mtv_tracker_jobs.get_job_name.call_args_list == [call('description')]
    assert mtv_tracker_jobs.make_precondition.call_args_list == [call('description_job')]
    assert mtv_tracker_jobs.common_job_args.call_args_list == [call(ignore_cache=True)]


class ImageUrl(str):
    def __new__(cls, *args, thumbnail=True, **kwargs):
        self = super().__new__(cls, *args, **kwargs)
        self._thumbnail = thumbnail
        return self

    @property
    def thumbnail_url(self):
        if self._thumbnail:
            return f'thumb_{self}'

@pytest.mark.parametrize(
    argnames='mediainfos_and_screenshots, exp_description',
    argvalues=(
        pytest.param(
            {
                'path/to/foo.mkv': {
                    'mediainfo': '<mediainfo for foo.mkv>',
                    'screenshot_urls': [ImageUrl('foo.1.png'), ImageUrl('foo.2.png'), ImageUrl('foo.3.png'), ImageUrl('foo.4.png')],
                },
            },
            (
                '[mediainfo]<mediainfo for foo.mkv>[/mediainfo]\n'
                '[center]'
                '[url=foo.1.png][img]thumb_foo.1.png[/img][/url]    [url=foo.2.png][img]thumb_foo.2.png[/img][/url]'
                '\n\n'
                '[url=foo.3.png][img]thumb_foo.3.png[/img][/url]    [url=foo.4.png][img]thumb_foo.4.png[/img][/url]'
                '[/center]'
                '\n'
                '[align=right][size=1]'
                f'Shared with [url={__homepage__}]{__project_name__}[/url]'
                '[/size][/align]'
            ),
            id='Single mkv',
        ),
        pytest.param(
            {
                'path/to/foo.mkv': {
                    'mediainfo': '',
                    'screenshot_urls': [ImageUrl('foo.1.png'), ImageUrl('foo.2.png'), ImageUrl('foo.3.png'), ImageUrl('foo.4.png')],
                },
            },
            (
                '[center]'
                '[url=foo.1.png][img]thumb_foo.1.png[/img][/url]    [url=foo.2.png][img]thumb_foo.2.png[/img][/url]'
                '\n\n'
                '[url=foo.3.png][img]thumb_foo.3.png[/img][/url]    [url=foo.4.png][img]thumb_foo.4.png[/img][/url]'
                '[/center]'
                '\n'
                '[align=right][size=1]'
                f'Shared with [url={__homepage__}]{__project_name__}[/url]'
                '[/size][/align]'
            ),
            id='Single mkv / Missing mediainfo',
        ),
        pytest.param(
            {
                'path/to/foo.1.mkv': {
                    'mediainfo': '<mediainfo for foo.1.mkv>',
                    'screenshot_urls': [ImageUrl('foo.11.png'), ImageUrl('foo.12.png'), ImageUrl('foo.13.png'), ImageUrl('foo.14.png')],
                },
                'path/to/foo.2.mkv': {
                    'mediainfo': '<mediainfo for foo.2.mkv>',
                    'screenshot_urls': [ImageUrl('foo.21.png'), ImageUrl('foo.22.png'), ImageUrl('foo.23.png'), ImageUrl('foo.24.png')],
                },
                'path/to/foo.3.mkv': {
                    'mediainfo': '<mediainfo for foo.3.mkv>',
                    'screenshot_urls': [ImageUrl('foo.31.png'), ImageUrl('foo.32.png'), ImageUrl('foo.33.png'), ImageUrl('foo.34.png')],
                },
            },
            (
                '[mediainfo]<mediainfo for foo.1.mkv>[/mediainfo]\n'
                '[center]'
                '[url=foo.11.png][img]thumb_foo.11.png[/img][/url]    [url=foo.12.png][img]thumb_foo.12.png[/img][/url]'
                '\n\n'
                '[url=foo.13.png][img]thumb_foo.13.png[/img][/url]    [url=foo.14.png][img]thumb_foo.14.png[/img][/url]'
                '[/center]'
                '\n[hr]\n'
                '[mediainfo]<mediainfo for foo.2.mkv>[/mediainfo]\n'
                '[center]'
                '[url=foo.21.png][img]thumb_foo.21.png[/img][/url]    [url=foo.22.png][img]thumb_foo.22.png[/img][/url]'
                '\n\n'
                '[url=foo.23.png][img]thumb_foo.23.png[/img][/url]    [url=foo.24.png][img]thumb_foo.24.png[/img][/url]'
                '[/center]'
                '\n[hr]\n'
                '[mediainfo]<mediainfo for foo.3.mkv>[/mediainfo]\n'
                '[center]'
                '[url=foo.31.png][img]thumb_foo.31.png[/img][/url]    [url=foo.32.png][img]thumb_foo.32.png[/img][/url]'
                '\n\n'
                '[url=foo.33.png][img]thumb_foo.33.png[/img][/url]    [url=foo.34.png][img]thumb_foo.34.png[/img][/url]'
                '[/center]'
                '\n'
                '[align=right][size=1]'
                f'Shared with [url={__homepage__}]{__project_name__}[/url]'
                '[/size][/align]'
            ),
            id='Multiple mkv',
        ),
        pytest.param(
            {
                'path/to/dvd/VIDEO_TS/VTS_01_0.IFO': {
                    'mediainfo': '<mediainfo for VTS_01_0.IFO>',
                    'screenshot_urls': [],
                },
                'path/to/dvd/VIDEO_TS/VTS_01_1.VOB': {
                    'mediainfo': '<mediainfo for VTS_01_1.VOB>',
                    'screenshot_urls': [ImageUrl('VTS_011.png'), ImageUrl('VTS_012.png'), ImageUrl('VTS_013.png'), ImageUrl('VTS_014.png')],
                },
            },
            (
                '[mediainfo]<mediainfo for VTS_01_0.IFO>[/mediainfo]\n'
                '[mediainfo]<mediainfo for VTS_01_1.VOB>[/mediainfo]\n'
                '[center]'
                '[url=VTS_011.png][img]thumb_VTS_011.png[/img][/url]    [url=VTS_012.png][img]thumb_VTS_012.png[/img][/url]'
                '\n\n'
                '[url=VTS_013.png][img]thumb_VTS_013.png[/img][/url]    [url=VTS_014.png][img]thumb_VTS_014.png[/img][/url]'
                '[/center]'
                '\n'
                '[align=right][size=1]'
                f'Shared with [url={__homepage__}]{__project_name__}[/url]'
                '[/size][/align]'
            ),
            id='Single VIDEO_TS',
        ),
        pytest.param(
            {
                'path/to/dvd/disc1/VIDEO_TS/VTS_01_0.IFO': {
                    'mediainfo': '<mediainfo for disc1/VTS_01_0.IFO>',
                    'screenshot_urls': [],
                },
                'path/to/dvd/disc1/VIDEO_TS/VTS_01_1.VOB': {
                    'mediainfo': '<mediainfo for disc1/VTS_01_1.VOB>',
                    'screenshot_urls': [ImageUrl('1/VTS_011.png'), ImageUrl('1/VTS_012.png'),
                                        ImageUrl('1/VTS_013.png'), ImageUrl('1/VTS_014.png')],
                },
            },
            (
                '[mediainfo]<mediainfo for disc1/VTS_01_0.IFO>[/mediainfo]\n'
                '[mediainfo]<mediainfo for disc1/VTS_01_1.VOB>[/mediainfo]\n'
                '[center]'
                '[url=1/VTS_011.png][img]thumb_1/VTS_011.png[/img][/url]    [url=1/VTS_012.png][img]thumb_1/VTS_012.png[/img][/url]'
                '\n\n'
                '[url=1/VTS_013.png][img]thumb_1/VTS_013.png[/img][/url]    [url=1/VTS_014.png][img]thumb_1/VTS_014.png[/img][/url]'
                '[/center]'
                '\n'
                '[align=right][size=1]'
                f'Shared with [url={__homepage__}]{__project_name__}[/url]'
                '[/size][/align]'
            ),
            id='Multiple VIDEO_TS',
        ),
    ),
)
@pytest.mark.asyncio
async def test_MtvTrackerJobs_generate_description(mediainfos_and_screenshots, exp_description, mtv_tracker_jobs, mocker):
    mocker.patch.object(type(mtv_tracker_jobs), 'mediainfos_and_screenshots', PropertyMock(
        return_value=mediainfos_and_screenshots,
    ))
    description = await mtv_tracker_jobs.generate_description()
    assert description == exp_description


def test_MtvTrackerJobs_mediainfo_from_all_videos(mtv_tracker_jobs):
    assert mtv_tracker_jobs.mediainfo_from_all_videos is False


def test_MtvTrackerJobs_screenshots_from_all_videos(mtv_tracker_jobs):
    assert mtv_tracker_jobs.screenshots_from_all_videos is False


def test_MtvTrackerJobs_post_data_autofill(mtv_tracker_jobs):
    assert mtv_tracker_jobs.post_data_autofill == {
        'submit': 'true',
        'MAX_FILE_SIZE': '2097152',
        'fillonly': 'auto fill',
        'category': '0',
        'Resolution': '0',
        'source': '12',
        'origin': '6',
        'title': '',
        'genre_tags': '---',
        'taglist': '',
        'autocomplete_toggle': 'on',
        'image': '',
        'desc': '',
        'fontfont': '-1',
        'fontsize': '-1',
        'groupDesc': '',
        'anonymous': '0',
    }


@pytest.mark.parametrize('anonymous, exp_anonymous', ((True, '1'), (False, '0')))
@pytest.mark.parametrize('ignore_dupes, exp_ignoredupes', ((True, '1'), (False, None)))
def test_MtvTrackerJobs_post_data_upload(anonymous, exp_anonymous, ignore_dupes, exp_ignoredupes,
                                         mtv_tracker_jobs, mock_job_attributes, mocker):
    mock_job_attributes(mtv_tracker_jobs)
    mocker.patch.object(mtv_tracker_jobs, 'get_job_attribute')
    mocker.patch.object(mtv_tracker_jobs, 'get_job_output')
    mocker.patch.object(type(mtv_tracker_jobs), 'options', PropertyMock(return_value={
        'anonymous': anonymous,
        'ignore_dupes': ignore_dupes,
    }))

    assert mtv_tracker_jobs.post_data_upload == {
        'submit': 'true',
        'category': mtv_tracker_jobs.get_job_attribute.return_value,
        'Resolution': '0',
        'source': '12',
        'origin': '6',
        'title': mtv_tracker_jobs.get_job_output.return_value,
        'genre_tags': '---',
        'autocomplete_toggle': 'on',
        'image': '',
        'desc': mtv_tracker_jobs.get_job_output.return_value,
        'fontfont': '-1',
        'fontsize': '-1',
        'groupDesc': '',
        'anonymous': exp_anonymous,
        'ignoredupes': exp_ignoredupes,
        'imdbID': mtv_tracker_jobs.get_job_output.return_value,
        # 'tmdbID': ...,
        # 'thetvdbID': ...,
        # 'tvmazeID': ...,
    }

    assert mtv_tracker_jobs.get_job_attribute.call_args_list == [call(mtv_tracker_jobs.category_job, 'choice')]
    assert mtv_tracker_jobs.get_job_output.call_args_list == [
        call(mtv_tracker_jobs.title_job, slice=0),
        call(mtv_tracker_jobs.description_job, slice=0),
        call(mtv_tracker_jobs.imdb_job, slice=0),
    ]


def test_MtvTrackerJobs_torrent_filepath(mtv_tracker_jobs, mock_job_attributes, mocker):
    mock_job_attributes(mtv_tracker_jobs)
    mocker.patch.object(mtv_tracker_jobs, 'get_job_output')

    assert mtv_tracker_jobs.torrent_filepath == mtv_tracker_jobs.get_job_output.return_value

    assert mtv_tracker_jobs.get_job_output.call_args_list == [
        call(mtv_tracker_jobs.create_torrent_job, slice=0),
    ]



import base64

import pytest

from upsies.trackers.nbl import NblTrackerConfig


@pytest.fixture
def tracker_config():
    return NblTrackerConfig()


def test_NblTrackerConfig_defaults(tracker_config):
    assert set(tracker_config) == {
        'upload_url',
        'announce_url',
        'apikey',
        'source',
        'randomize_infohash',
        'exclude',

        # Inherited from TrackerConfigBase
        'add_to',
        'copy_to',
    }


def test_NblTrackerConfig_defaults_base_url(tracker_config):
    assert tracker_config['upload_url'] == base64.b64decode('aHR0cHM6Ly9uZWJ1bGFuY2UuaW8vdXBsb2FkLnBocA==').decode('ascii')


def test_NblTrackerConfig_defaults_password(tracker_config):
    assert tracker_config['apikey'] == ''
    assert tracker_config['apikey'].description == (
        'Your personal private API key.\n'
        'Get it from the website: <USERNAME> -> Settings -> API keys'
    )


def test_NblTrackerConfig_defaults_announce_url(tracker_config):
    assert tracker_config['announce_url'] == ''
    assert tracker_config['announce_url'].description == (
        'The complete announce URL with your private passkey.\n'
        'Get it from the website: Shows -> Upload -> Your personal announce URL'
    )


def test_NblTrackerConfig_defaults_source(tracker_config):
    assert tracker_config['source'] == 'NBL'


def test_NblTrackerConfig_defaults_exclude(tracker_config):
    assert tracker_config['exclude'] == ()


def test_NblTrackerConfig_argument_definitions(tracker_config):
    assert tracker_config.argument_definitions == {}
